@@ -1,6 +1,6 @@
 package Tcl;
 
-$Tcl::VERSION = '1.02_50';
+$Tcl::VERSION = '1.51_01';
 
 =head1 NAME
 
@@ -74,7 +74,16 @@ returned (the object argument omitted in each case).
 
 =item $interp->Init ()
 
-Invoke I<Tcl_Init> on the interpeter.
+Invoke I<Tcl_Init> on the interpreter.
+
+=item $interp->CreateSlave (NAME, SAFE)
+
+Invoke I<Tcl_CreateSlave> on the interpreter.  Name is arbitrary.
+The safe variable, if true, creates a safe sandbox interpreter.
+ See: http://www.tcl.tk/software/plugin/safetcl.html
+      http://www.tcl.tk/man/tcl8.4/TclCmd/safe.htm
+
+This command returns a new interpreter.
 
 =item $interp->Eval (STRING, FLAGS)
 
@@ -90,7 +99,7 @@ Tcl::EVAL_GLOBAL or Tcl::EVAL_DIRECT.
 
 =item $interp->GlobalEval (STRING)
 
-REMOVED.  Evalulate script STRING at global level.
+REMOVED.  Evaluate script STRING at global level.
 Call I<Eval>(STRING, Tcl::EVAL_GLOBAL) instead.
 
 =item $interp->EvalFile (FILENAME)
@@ -121,7 +130,7 @@ created using CreateCommand subroutine (see below).
 2.  All references to scalars will be substituted with names of Tcl variables
 transformed appropriately.
 
-These first two items allows to write and expect it to work properly such
+These first two items allow one to write and expect it to work properly such
 code as:
 
   my $r = 'aaaa';
@@ -152,6 +161,8 @@ example this could be variable associated with a widget.
 
 deletes and returns a reference corresponding to NAME, which was associated
 during previously called C<< $interpnt->call(...) >> preprocessing.
+this follows _code_dispose about if NAME is a TClNAME or a DESCNAME.
+if NAME is a VARNAME then it is just deleted and returned.
 
 =item $interp->icall (PROC, ARG, ...)
 
@@ -174,7 +185,7 @@ spaces embedded in any ARG will not cause it to be split into two Tcl
 arguments before being passed to PROC.  This differs from icall/call in
 that it directly invokes the command name without allowing for command
 tracing or making use of Tcl's unknown (AUTOLOAD) mechanism.  If the
-command does not already exist in the interpreter, and error will be
+command does not already exist in the interpreter, an error will be
 thrown.
 
 Arguments are converted efficiently from Perl SVs to Tcl_Objs.  A Perl AV
@@ -211,7 +222,7 @@ deleteProc field presumably obtained in a similar way.
 (either a sub name, a sub reference or an anonymous sub). In this case
 CLIENTDATA can be any perl scalar (e.g. a ref to some other data) and
 DELETEPROC must be a perl sub too. When CMDNAME is invoked in the Tcl
-interpeter, the arguments passed to the Perl sub CMDPROC are
+interpreter, the arguments passed to the Perl sub CMDPROC are
 
     (CLIENTDATA, INTERP, LIST)
 
@@ -221,7 +232,7 @@ If the 1-bit of FLAGS is set then the 3 first arguments on the call
 to CMDPROC are suppressed.
 As usual in Tcl, the first element of the list is CMDNAME itself.
 When CMDNAME is deleted from the interpreter (either explicitly with
-I<DeleteCommand> or because the destructor for the interpeter object
+I<DeleteCommand> or because the destructor for the interpreter object
 is called), it is passed the single argument CLIENTDATA.
 
 =item $interp->DeleteCommand (CMDNAME)
@@ -232,6 +243,83 @@ this point. When a Tcl interpreter object is destroyed either explicitly
 or implicitly, an implicit I<DeleteCommand> happens on all its currently
 registered commands.
 
+
+=item (TCLNAME,GENCODE) = $interp->create_tcl_sub(CODEREF, EVENTS, TCLNAME, DESCRNAME)
+
+Creates a COMMAND called TCLNAME calling CODEREF in the interpreter $interp,
+and adds it to the internal tracking structure.
+DESCRNAME and TCLNAME should not begin
+with C<=>
+or C<@>
+or C<::perl::>
+to avoid collisions.
+If TCLNAME IS blank or undef, it is constructed from the CODEREF address.
+GENCODE starts as TCLNAME but gets @$EVENTS which can contain %vars joined to it.
+
+TCLNAME and DESCRNAME get stored in an internal structure,
+and can be used to purge things from the command table via code_destroy or $interp->delete_ref;
+
+Returns (TCLNAME,GENCODE).
+if you are creating code refs with this you can continue to use the same coderef and it will be converted on each call.
+but if you save GENCODE, you can replace the anon-coderef call in the tcl command with GENCODE.
+
+for instance
+
+  $interp->call('FILEEVENT',$fileref,WRITABLE=>sub {...});
+
+can be replaced by
+
+  my ($tclcode,$gencode)=$interp->create_tcl_sub(sub{...}, EVENTS, TCLNAME, DESCRNAME);
+  $interp->call('FILEEVENT',$gencode,WRITABLE=>$gencode);
+
+or
+
+  my $sub=sub{....};
+  $interp->call('FILEEVENT',$fileref,WRITABLE=>$sub);
+
+can be replaced by
+
+  my ($tclcode,$gencode)=$interp->create_tcl_sub($sub, EVENTS, TCLNAME, DESCRNAME);
+  $interp->call('FILEEVENT',$gencode,WRITABLE=>$gencode);
+
+although
+
+  $interp->call('FILEEVENT',$fileref,WRITABLE=>$sub);
+
+will still work fine, too.
+
+Then you later call
+
+  $interp->delete_ref($tclname);
+
+when you are finished with that sub to clean it from the internal tracking and command table.
+This means no automatic cleanup will occur on the sub{...} or $sub
+
+
+And after the destroy inside Tcl any triggering writable on $fileref will fail as well.
+so it should be replaced first via
+  $interp->call('FILEEVENT',$fileref,WRITABLE=>'');
+
+=item (CODEREF) = Tcl::_code_dispose(NAME)
+	
+Purges the internal table of a NAME
+and may initiate destruction of something created thru call or create_tcl_sub
+
+TCLNAME and DESCRNAME get stored in an internal structure,
+and can be used to purge things form the command table.
+calling _code_dispose on a TCLNAME returned from create_tcl_sub removes all use instances and purges the command table.
+calling _code_dispose on a DESCRNAME passed to create_tcl_sub removes only that instance
+Code used in a DESCRNAME may be used in other places as well,
+only when the last usage is purged does the entry get purged from the command table
+
+While the internal tracking structure saves the INTERP the code was added to,
+it itself does not keep things separated by INTERP,
+A TCLNAME or DESCRNAME can only exist in one INTERP at a time,
+using a new INTERP just causes the one in the last INTERP to disappear,
+and probably end up with the Tcl code getting deleted
+
+Returns (CODEREF), this is the original coderef
+
 =item $interp->SetResult (STRING)
 
 Sets Tcl interpreter result to STRING.
@@ -287,6 +375,10 @@ behaves as in I<SetVar> above.
 Unsets the element VARNAME1(VARNAME2) of a Tcl array.
 The optional argument FLAGS behaves as in I<SetVar> above.
 
+=item $interp->InterpDeleted ()
+
+See Tcl C API documentation for I<Tcl_InterpDeleted>().
+
 =back
 
 =head2 Linking Perl and Tcl variables
@@ -325,70 +417,39 @@ namespace C< ::perl > following objects are created:
 
 So it is possible to use Perl objects from within Tcl.
 
-=head2 Moving Tcl/Tk around with Tcl.pm
-
-NOTE: explanations below is for developers managing Tcl/Tk installations
-itself, users should skip this section.
-
-In order to create Tcl/Tk application with this module, you need to make
-sure that Tcl/Tk is available within visibility of this module. There are
-many ways to achieve this, varying on ease of starting things up and
-providing flexible moveable archived files.
-
-Following list enumerates them, in order of increased possibility to change
-location.
-
-=over
-
-=item *
-
-First method
-
-Install Tcl/Tk first, then install Perl module Tcl, so installed Tcl/Tk will
-be used. This is most normal approach, and no care of Tcl/Tk distribution is
-taken on Perl side (this is done on Tcl/Tk side)
-
-=item *
-
-Second method
-
-Copy installed Tcl/Tk binaries to some location, then install Perl module Tcl
-with a special action to make Tcl.pm know of this location. This approach
-makes sure that only chosen Tcl installation is used.
-
-=item *
-
-Third method
-
-During compiling Tcl Perl module, Tcl/Tk could be statically linked into
-module's shared library and all other files zipped into a single archive, so
-each file extracted when needed.
-
-To link Tcl/Tk binaries, prepare their libraries and then instruct Makefile.PL
-to use these libraries in a link stage.
-(TODO provide better detailed description) 
-
-=back
-
 =cut
 
 use strict;
 
+sub TRACE_SHOWCODE(){0}      # display generated code in call();
+sub TRACE_CREATECOMMAND(){0} # display sub creates;
+sub TRACE_DELETECOMMAND(){0} # display sub deletes;
+
+sub SAVEALLCODES () {1}    # simulate the v1.05 way of saving only last call
+sub SAVENOCODE() {1}       # simulate the v1.05 way of leaving any existing
+                           #  $anon_refs{$descrname} alone if new line has no coderefs
+
 our $DL_PATH;
 unless (defined $DL_PATH) {
     $DL_PATH = $ENV{PERL_TCL_DL_PATH} || $ENV{PERL_TCL_DLL} || "";
 }
 
-unless ($DL_PATH) {
-    require Config;
-    for my $inc (@INC) {
-        my $tkkit = "$inc/auto/Tcl/tkkit.$Config::Config{so}";
-        if (-f $tkkit) {
-            $DL_PATH = $tkkit;
-            last;
+=for ignore
+sub Tcl::seek_tkkit {
+    # print STDERR "wohaaa!\n";
+    unless ($DL_PATH) {
+        require Config;
+        for my $inc (@INC) {
+            my $tkkit = "$inc/auto/Tcl/tkkit.$Config::Config{so}";
+            if (-f $tkkit) {
+                $DL_PATH = $tkkit;
+                last;
+            }
         }
     }
 }
+=cut
+seek_tkkit() if defined &seek_tkkit;
 
 
 my $path;
@@ -413,8 +474,10 @@ sub new {
     return $int;
 }
 
+my $pid = $$; # see rt ticket 77522
 END {
-    Tcl::_Finalize();
+    Tcl::_Finalize()
+        if $$ == $pid;
 }
 
 # %anon_refs keeps track of anonymous subroutines and scalar/array/hash
@@ -425,6 +488,32 @@ END {
 # subs and with 'tie' for other)
 
 my %anon_refs;
+sub _anon_refs_cheat { return \%anon_refs;}
+
+# Coderef tracking creates 3 types of entries in %anon_refs
+# 1) Tcl::Code
+#   this is added to %anon_refs when a coderef is found in a tcl call
+#           [\$sub, $interp, $tclname,$descrname]
+#     It is an array with four elements stored at the key $descrname
+#       where [0] $sub is the perlcoderef,
+#             [1] $interp is the tcl interpreter it is in,
+#             [2] $tclname is the name for the coderef in tcl
+#                 and serves as a pointer to the Tcl::Cmdbase entry
+#             [3] $descrname is the name passed from call() or the app
+# 2) Tcl::Cmdbase
+#   this is added to %anon_refs when a new code reference is added to the tcl internal tables.
+#     It is an array with two elements stored at the key $tclname
+#     The first element[0] is an array [\$sub, $interp, $tclname,0]
+#       where [0] $sub is the perlcoderef,
+#             [1] $interp is the tcl interpreter it is in,
+#             [2] $tclname is the name for the coderef in tcl
+#             [3] is the number of Tcl::Code elements that reference it.
+#     The second element [1] is a hash
+#      for each Tcl::Code $descrname there is a $descrname key here,
+#        the value at the key is the number of Tcl::Code of that $descrname pointing to this Tcl::Cmdbase
+#         when this is >1 it is because there are Tcl::Code elements in $lastcodes that will be destroyed soon
+#          and we want to be able to retain the $descrname backlink after they are destroyed
+# 3) an array of Tcl::Code elements stored at the key $descrname
 
 # (TODO -- find out how to check for refcounting and proper releasing of
 # resources)
@@ -435,9 +524,23 @@ my %anon_refs;
 sub call {
     my $interp = shift;
     my @args = @_;
-    my $current_r = join ' ', grep {defined} grep {!ref} @args;
+    # add = so  as not to interfere with user defined DESCRNAMEs
+    my $descrname = '='.join ' ', grep {defined} grep {!ref} @args;
     my @codes;
 
+    my $lastcodes = $anon_refs{$descrname}; # could be an array, want to hold all for now so they can be reused
+    unless  (SAVENOCODE()) {
+      unless  ($#args == 1 and $args[0] eq 'set') {
+          # this was to clean up things like
+          # Tkx::fileevent($remote, writable  => sub {...});
+          # when this was run later
+          # Tkx::fileevent($fh, writable  => '');
+          # but "set foo" doesnt clear its setting, it returns its value
+          # after busting the test in the Tkx tests for set foo
+          #  i decided to disable this for now
+          delete $anon_refs{$descrname};        # if old one had a call and new one doesnt make sure it goes away
+          }
+    }
     # Process arguments looking for special cases
     for (my $argcnt=0; $argcnt<=$#args; $argcnt++) {
 	my $arg = $args[$argcnt];
@@ -446,8 +549,8 @@ sub call {
 	if ($ref eq 'CODE' || $ref eq 'Tcl::Code') {
 	    # We have been passed something like \&subroutine
 	    # Create a proc in Tcl that invokes this subroutine (no args)
-	    $args[$argcnt] = $interp->create_tcl_sub($arg, undef, undef, $current_r);
-	    push @codes, $anon_refs{$current_r}; # push CODE also only to keep it from early disposal
+	    $args[$argcnt] = $interp->create_tcl_sub($arg, undef, undef, $descrname);
+	    push @codes, $anon_refs{$descrname}; # push CODE also only to keep it from early disposal
 	}
 	elsif ($ref eq 'SCALAR') {
 	    # We have been passed something like \$scalar
@@ -498,25 +601,25 @@ sub call {
 	    $args[$argcnt] =
 		$interp->create_tcl_sub(sub {
 		    $arg->[0]->(@_, @$arg[1..$#$arg]);
-		}, $events, undef, $current_r);
-	    push @codes, $anon_refs{$current_r};
+		}, $events, undef, $descrname);
+	    push @codes, $anon_refs{$descrname};
 	}
 	elsif ($ref eq 'REF' and ref($$arg) eq 'SCALAR') {
 	    # this is a very special shortcut: if we see construct like \\"xy"
 	    # then place proper Tcl::Ev(...) for easier access
 	    my $events = [map {"%$_"} split '', $$$arg];
-	    if (ref($args[$argcnt+1]) eq 'ARRAY' && 
+	    if (ref($args[$argcnt+1]) eq 'ARRAY' &&
 		ref($args[$argcnt+1]->[0]) eq 'CODE') {
 		$arg = $args[$argcnt+1];
 		$args[$argcnt] =
 		    $interp->create_tcl_sub(sub {
 			$arg->[0]->(@_, @$arg[1..$#$arg]);
-		    }, $events, undef, $current_r);
-		push @codes, $anon_refs{$current_r};
+		    }, $events, undef, $descrname);
+		push @codes, $anon_refs{$descrname};
 	    }
 	    elsif (ref($args[$argcnt+1]) eq 'CODE') {
-		$args[$argcnt] = $interp->create_tcl_sub($args[$argcnt+1],$events, undef, $current_r);
-		push @codes, $anon_refs{$current_r};
+		$args[$argcnt] = $interp->create_tcl_sub($args[$argcnt+1],$events, undef, $descrname);
+		push @codes, $anon_refs{$descrname};
 	    }
 	    else {
 		warn "not CODE/ARRAY expected after description of event fields";
@@ -525,27 +628,50 @@ sub call {
 	}
     }
 
+    $lastcodes = []; # now let any from last use be destroyed
+
+    showcode(\@args) if TRACE_SHOWCODE();
+
     if ($#codes>-1 and $args[0] eq 'after') {
-	if ($args[1] =~ /^\d+$/) {
+#	AFTERS can clog up the tables real quick, so plan to delete them via Tcl::Code::DESTROY
+	my $delay='';
+	if ($args[1] =~ /^\d+$/)   { $delay=$args[1]+1000; }
+	elsif ($args[1] eq 'idle') { $delay='idle'; } # just hope they run in order
+        if ($delay) {
 	    my $id = $interp->icall(@args);
 	    #print STDERR "rebind for $interp;$id\n";
 	    # in 'after' methods, disposal of CODE REFs based on 'after' id
 	    # i.e based on return value of tcl call
-	    $anon_refs{"$interp;$id"} = \@codes;
-	    delete $anon_refs{$current_r};
-	    # plan deleting that entry, hence Tcl command during Tcl::Code::DESTROY
-	    # TODO - this +1000 is wrong... should
-	    #$interp->invoke('after',$args[1]+1000, "perl::Eval {Tcl::_code_dispose('$interp;$id')}");
-	    #$interp->invoke('trace','add','variable',$id,'unset',"puts {deleting $id}");
-	    #$interp->invoke('trace','add','variable',$id,'read',"puts {deleting $id}");
-	    #$interp->invoke('trace','add','variable',$id,'write',"puts {deleting $id}");
-	    #$interp->invoke('trace','add','variable',$id,'qrite',"puts {deleting $id}");
+	    my $newname='@'.$interp.';'.$id;
+	    $anon_refs{$newname} = \@codes;
+	    for my $code (@codes) {
+              my $ridptr=$anon_refs{$code->[2]}[1];
+              $ridptr->{$newname}++ ;    # save under new name
+              $ridptr->{$descrname}-- ;  # maybe take out this name
+              unless ($ridptr->{$descrname}>0){
+                delete $ridptr->{$descrname}; # clean up old
+              }
+            }
+	    delete $anon_refs{$descrname};                                    # and now its clean
+	    for my $code (@codes) { $code->[3]=$newname;                    } # save under new name
+	    # this should trigger DESTROY unless a newer after or something else still holds a $tclname in the list
+	    $interp->invoke('after',$delay, "perl::Eval {Tcl::_code_dispose('$newname')}");
+	    showcode(      ['after',$delay, "perl::Eval {Tcl::_code_dispose('$newname')}"]   ) if TRACE_SHOWCODE();
 	    return $id;
-	} elsif ($args[1] eq 'idle') {
-	    # no planned CODE REF disposal, just do as is
-	    return $interp->icall(@args);
-	}
+        }   # delay
 	# if we're here - user does something wrong, but there is nothing we worry about
+    } # codes and after
+
+
+    # got to keep all of them alive , not just the last
+    # incase there are lines that have more than one  like ===== fileevent readable=>$sub1 writable=>$sub2
+    # although that isnt legal, but this is                ===== if 1 $sub1 $sub2
+    #  the downside is that add/delete processing goes on for $sub1 every call if we only keep the last
+    if (SAVEALLCODES()) {
+        if (scalar(@codes)>1) {  # no reason to save an array of just 1
+            delete $anon_refs{$descrname};
+            $anon_refs{$descrname}=\@codes;
+        }
     }
 
     # Done with special var processing.  The only processing that icall
@@ -558,13 +684,15 @@ sub call {
     #    return $interp->icall(@args);
     # a bit of complications only to allow stack trace, i.e. in case of errors
     # user will get error pointing to his program and not in this module.
+    # and also 'after' tcl method makes bit harder
 
     if (wantarray) {
 	my @res;
 	eval { @res = $interp->icall(@args); };
 	if ($@) {
+	    my $errmsg = $@;     # 'require Carp' might destroy $@;
 	    require Carp;
-	    Carp::confess ("Tcl error '$@' while invoking array result call:\n" .
+	    Carp::croak ("Tcl error '$errmsg' while invoking array result call:\n" .
 		"\t\"@args\"");
 	}
 	return @res;
@@ -572,14 +700,19 @@ sub call {
 	my $res;
 	eval { $res = $interp->icall(@args); };
 	if ($@) {
+	    my $errmsg = $@;     # 'require Carp' might destroy $@;
 	    require Carp;
-	    Carp::confess ("Tcl error '$@' while invoking scalar result call:\n" .
+	    Carp::croak ("Tcl error '$errmsg' while invoking scalar result call:\n" .
 		"\t\"@args\"");
 	}
 	return $res;
     }
 }
 
+sub showcode{
+  print 'TCL::TRACE_SHOWCODE:'.join(' ',@{$_[0]})."\n";
+}
+
 # create_tcl_sub will create TCL sub that will invoke perl CODE ref
 # If $events variable is specified then special processing will be
 # performed to provide needed '%' variables.
@@ -587,38 +720,123 @@ sub call {
 # otherwise it will have machine-readable name.
 # Returns tcl script suitable for using in tcl events.
 sub create_tcl_sub {
-    my ($interp,$sub,$events,$tclname, $rname) = @_;
+    my ($interp,$sub,$events,$tclname, $descrname) = @_;
+    # rnames and tclnames beginning = or @ or ::perl:: are reserved for internal use
     unless ($tclname) {
 	# stringify sub, becomes "CODE(0x######)" in ::perl namespace
 	$tclname = "::perl::$sub";
     }
 
-    #print STDERR "...=$rname\n";
-    $interp->CreateCommand($tclname, $sub, undef, undef, 1);
+    #print STDERR "...=$descrname\n";
 
-    # following line a bit more tricky than it seems to.
-    # because the whole intent of the %anon_refs hash is to have refcount
+    # the following is a bit more tricky than it seems to.
+    # because the whole intent of the Tcl::Cmdbase entries in %anon_refs hash is to have refcount
     # of (possibly) anonymous sub that is happen to be passed,
     # and, if passed for the same widget but arguments are same - then
-    # previous instance will be overwriten, and sub will be destroyed due
-    # to reference count, and Tcl method will also be destroyed during
-    # Tcl::Code::DESTROY
-    $anon_refs{$rname} = bless [\$sub, $interp], 'Tcl::Code';
+    # previous instance will be overwritten, and sub will be destroyed due
+    # to no reference count, and command table entry will also be destroyed during
+    # Tcl::Cmdbase::DESTROY
+
+    unless (exists $anon_refs{$tclname}) {
+      $anon_refs{$tclname} = bless (
+                                    [[\$sub, $interp, $tclname,0]
+                                    ,{} # Tcl::Codes register here so they can get deleted
+                                    ],'Tcl::Cmdbase');
+      $interp->CreateCommand($tclname, $sub, undef, undef, 1);
+
+      print "TCL::TRACE_CREATECOMMAND: $interp -> $descrname ( $tclname => $sub ,undef,undef,1 )\n" if TRACE_CREATECOMMAND();
+      delete $anon_refs{$descrname};
+    }
+    my @newtcl=@{$anon_refs{$tclname}[0]};
+    pop @newtcl;
+    push @newtcl,$descrname;
+
+    bless( \@newtcl, 'Tcl::Code');
+    $anon_refs{$descrname} = \@newtcl;
+    $anon_refs{$tclname}[0][3]++;          # push ref ct
+    $anon_refs{$tclname}[1]{$descrname}++; #register as user
+
+    my $gencode = $tclname;
 
     if ($events) {
 	# Add any %-substitutions to callback
-	$tclname = "$tclname " . join(' ', @{$events});
+	$gencode = "$tclname " . join(' ', @{$events});
     }
-    return $tclname;
+    return $tclname , $gencode; # return both the base and code-call
 }
 
 sub _code_dispose {
     my $k = shift;
-    #print STDERR "_code_dispose $k\n";
-    #my $int = $anon_refs{$k}->[0]->[1];
-    #my @r = $int->Eval("after info $id"); # why do not work?
-    #print STDERR "r=@r\n";
-    delete $anon_refs{$k};
+    return undef unless(exists $anon_refs{$k});
+    my $ret = undef;
+    my $ref = ref($anon_refs{$k});
+    if    ($ref eq 'Tcl::Code') {
+        $ret = $anon_refs{$k}[0];
+        delete $anon_refs{$k};
+    }
+    elsif ($ref eq 'Tcl::Cmdbase') {
+       ## mainly for calling from exterior program when it is one that calls ($tclname)=create_tcl_sub()..  too
+       my $atcl = $anon_refs{$k};
+       return undef unless ($atcl);
+       $ret = $atcl->[0][0];
+       my @keys=keys(%{$atcl->[1]});
+       for my $key (@keys) {
+         _code_dispose($key); # last one deletes me
+       }
+       delete $anon_refs{$k}; # but be sure
+    }
+    elsif ($ref eq 'ARRAY') {
+      my $atclarray = $anon_refs{$k};
+      $ret = $atclarray->[0][0];  # eh, return the first one, prob discarded anyway
+      delete $anon_refs{$k}; # bunch of Tcl::Code
+      }
+    return $$ret;  # original delete_ref result, prob just discarded
+}
+
+sub delete_ref {
+    my $interp = shift;
+    my $name = shift;
+    my $iam = $anon_refs{$name};
+    my $ref = ref($iam);
+    if    ($ref eq 'ARRAY')       { return _code_dispose ($name);}
+    elsif ($ref eq 'Tcl::Code')   { return _code_dispose ($name);}
+    elsif ($ref eq 'Tcl::Cmdbase'){ return _code_dispose ($name);}
+    else { # these are the ties ??
+      $interp->UnsetVar($name); #TODO: will this delete variable in Tcl?
+      delete $anon_refs{$name};
+      untie $$iam;
+      return $iam;
+      }
+}
+
+sub return_ref {
+    my $interp = shift;
+    my $name = shift;
+    my $iam = $anon_refs{$name};
+    my $ref = ref($iam);
+    if    ($ref eq 'ARRAY')       { return ${$iam->[0][0]};}  # gotta pick one
+    elsif ($ref eq 'Tcl::Code')   { return ${$iam->[0]};}
+    elsif ($ref eq 'Tcl::Cmdbase'){ return ${$iam->[0][0]};}
+    return $anon_refs{$name};
+}
+
+sub _code_clear {
+  # for testing
+  my $debug = shift;
+  print "_code_clear ARRAY\n"          if ($debug);
+  for my $kk (keys %anon_refs) {
+    if (ref($anon_refs{$kk}) eq 'ARRAY'){ print "ARRAY $kk\n"        if ($debug); delete $anon_refs{$kk};}
+    }
+
+  print "_code_clear Tcl::Code list\n" if ($debug);
+  for my $kk (keys %anon_refs) {
+    if (ref($anon_refs{$kk}) eq 'Tcl::Code'){ print "Code $kk\n"     if ($debug); delete $anon_refs{$kk};}
+    }
+
+  print "_code_clear Tcl::Cmdbase\n"   if ($debug);
+  for my $kk (keys %anon_refs) {
+    if (ref($anon_refs{$kk}) eq 'Tcl::Cmdbase'){ print "Code $kk\n"  if ($debug); delete $anon_refs{$kk};}
+    }
 }
 
 
@@ -630,17 +848,50 @@ sub Ev {
 
 package Tcl::Code;
 
-# only purpose is to track CODE REFs passed to 'call' method 
+# purpose is to track CODE REFs still "in use"
 # (often these are anon subs)
-# so to bless it to this package and then catch deleting it, so 
+
+# each has ptr to the Tcl::Cmdbase entry that tracks the command table
+# so to bless it to this package and then catch deleting it, so
 # to do cleaning up
 
 sub DESTROY {
-    my $rsub = $_[0]->[0];
-    my $interp = $_[0]->[1];
-    my $tclname = "::perl::$$rsub";
-    #print STDERR "CODE::DESTROY[[@_]] $tclname\n";
-    $interp->DeleteCommand($tclname) if 0 and defined $tclname;  #TODO-ed
+#    my $rsub    = $_[0]->[0]; # don't really need it here anymore,
+#    my $interp  = $_[0]->[1];
+
+    my $tclname     = $_[0]->[2];
+    my $descrname   = $_[0]->[3];
+    my $tclptr = $anon_refs{$tclname}->[0];
+
+    return unless ($tclptr);                     # should exist but be safe
+    $tclptr->[3]--;
+    unless($tclptr->[3]>0) {
+      delete $anon_refs{$tclname};               # kill the Cmdbase
+    }
+    else {
+      $anon_refs{$tclname}->[1]{$descrname}--;
+      unless ( $anon_refs{$tclname}->[1]{$descrname}>0) {
+          delete $anon_refs{$tclname}->[1]{$descrname};  # unregister me
+      }
+    }
+}
+
+package Tcl::Cmdbase;
+
+# only purpose is to track CODE REFs passed to 'call' method
+# (often these are anon subs)
+# so to bless it to this package and then catch deleting it, so
+# to do cleaning up
+
+sub DESTROY {
+# $_[0] points to a [Tcl::code,{$descrname=>1,...}] set
+    my $interp  = $_[0]->[0]->[1];
+    my $tclname = $_[0]->[0]->[2];
+    return unless ($tclname);
+    if (defined $interp) {
+      $interp->DeleteCommand($tclname);
+      print "TCL::TRACE_DELETECOMMAND: $interp -> ( $tclname )\n" if Tcl::TRACE_DELETECOMMAND();
+    }
 }
 
 package Tcl::List;
@@ -677,24 +928,22 @@ sub FIRSTKEY {
 	unless @{$obj} == 2 || @{$obj} == 3;
     my ($interp, $varname, $flags) = @$obj;
     $arraystates{$varname} = $interp->invoke("array","startsearch",$varname);
-    my $r = $interp->invoke("array","nextelement",$varname,$arraystates{$varname});
-    if ($r eq '') {
-	delete $arraystates{$varname};
-	return undef;
-    }
-    return $r;
+    return $obj->_next_element(@_);
 }
 sub NEXTKEY {
     my $obj = shift;
     die "STORE Usage: objdata @{$obj} $#{$obj}, not 2 or 3 (@_)"
 	unless @{$obj} == 2 || @{$obj} == 3;
+    return $obj->_next_element(@_);
+}
+sub _next_element {
+    my $obj = shift;
     my ($interp, $varname, $flags) = @$obj;
-    my $r = $interp->invoke("array","nextelement",$varname,$arraystates{$varname});
-    if ($r eq '') {
+    unless ($interp->invoke('array','anymore',$varname,$arraystates{$varname})) {
 	delete $arraystates{$varname};
 	return undef;
     }
-    return $r;
+    return $interp->invoke('array','nextelement',$varname,$arraystates{$varname});
 }
 sub CLEAR {
     my $obj = shift;
@@ -707,8 +956,9 @@ sub CLEAR {
 sub DELETE {
     my $obj = shift;
     unless (@{$obj} == 2 || @{$obj} == 3) {
+	my @args = @_;
 	require Carp;
-	Carp::croak("STORE Usage: objdata @{$obj} $#{$obj}, not 2 or 3 (@_)");
+	Carp::croak("STORE Usage: objdata @{$obj} $#{$obj}, not 2 or 3 (@args)");
     }
     my ($interp, $varname, $flags) = @{$obj};
     my ($str1) = @_;
@@ -798,24 +1048,24 @@ An example:
 
   use strict;
   use Tcl;
-  
-  my $int = new Tcl;
-  
+
+  my $int = Tcl->new;
+
   $tcl::foo = 'qwerty';
   $int->export_to_tcl(subs_from=>'tcl',vars_from=>'tcl');
-  
+
   $int->Eval(<<'EOS');
   package require Tk
-  
+
   button .b1 -text {a fluffy button} -command perl::fluffy_sub
   button .b2 -text {a foo button} -command perl::foo
   entry .e -textvariable perl::foo
   pack .b1 .b2 .e
   focus .b2
-  
+
   tkwait window .
   EOS
-  
+
   sub tcl::fluffy_sub {
       print "Hi, I am a fluffy sub\n";
   }
@@ -875,7 +1125,7 @@ sub export_to_tcl {
 		local $_ = ${"$vars_from\::$name"};
 		tie ${"$vars_from\::$name"}, 'Tcl::Var', $int, "$tcl_namespace$name";
 		${"$vars_from\::$name"} = $_;
-	    } 
+	    }
 	    if (0) {
 		# array, hash - no need to do anything.
 		# (or should we?)
@@ -915,10 +1165,13 @@ sub export_tcl_namespace {
 
 =head1 AUTHORS
 
- Malcolm Beattie, mbeattie@sable.ox.ac.uk, 23 Oct 1994.
- Vadim Konovalov, vkon@cpan.org, 19 May 2003.
- Jeff Hobbs, jeff (a) activestate . com, 22 Mar 2004.
- Gisle Aas, gisle (a) activestate . com, 14 Apr 2004.
+ Malcolm Beattie, 23 Oct 1994
+ Vadim Konovalov, 19 May 2003
+ Jeff Hobbs, jeff (a) activestate . com, 22 Mar 2004
+ Gisle Aas, gisle (a) activestate . com, 14 Apr 2004
+
+Special thanks for contributions to Jan Dubois, Slaven Rezic, Paul Cochrane,
+Huck Finn, Christopher Chavez, SJ Luo.
 
 =head1 COPYRIGHT
 
@@ -930,4 +1183,3 @@ See http://www.perl.com/perl/misc/Artistic.html
 =cut
 
 1;
-
